from enum import Enum

class ViralLoads(Enum):
    COVID_OVERALL = "Ref: Viral load - covid overal viral load data"
    SYMPTOMATIC_FREQUENCIES = "Ref: Viral load - symptomatic viral load frequencies"
